from functools import lru_cache
from typing import Dict, List, Optional, Tuple, Type

from lxml import etree
from lxml.etree import Element

from scenario_gym.catalog_entry import Catalog, CatalogEntry
from scenario_gym.entity import Entity, Pedestrian, Vehicle

DEFAULT_ENTITY_TYPES = (Vehicle, Pedestrian)


def load_object(
    catalog: Catalog,
    entry: Element,
    entity_types: List[Type[Entity]],
    catalog_objects: List[Type[CatalogEntry]],
) -> Optional[Entity]:
    """Try to load a catalog entry with given catalog objects."""
    for Ent, Obj in zip(entity_types, catalog_objects):
        types = Obj.xosc_names if Obj.xosc_names is not None else [Obj.__name__]
        if entry.tag in types:
            obj = Obj.from_xml(catalog, entry)
            return Ent(obj)


@lru_cache(maxsize=None)
def read_catalog(
    catalog_file: str,
    relative_catalog_path: Optional[str] = None,
    entity_types: Optional[Tuple[Type[Entity]]] = None,
) -> Tuple[Catalog, Dict[str, Entity]]:
    """
    Read a catalog and return it's name and a dictionary of entities.

    Parameters
    ----------
    catalog_file : str
        Filepath of the catalog file.

    relative_catalog_path : Optional[str]
        Relative path used by OpenSCENARIO files when loading this catalog.

    entity_types : Optional[Tuple[Type[CatalogObject]]]
        Tuple of extra subclasses of CatalogObject that will be used when reading
        catalogs. Can be used for reading custom objects from catalog files. Must
        be immutable or lru_cache

    Returns
    -------
    catalog : Catalog
        The catalog object.

    entries : Dict[str, Entity]
        A dictionary mapping each entry name to an entity with that catalog entry.
        These can then be cloned to create entities using the chosen catalog entry.

    """
    if entity_types is None:
        entity_types = DEFAULT_ENTITY_TYPES
    else:
        entity_types = entity_types + DEFAULT_ENTITY_TYPES

    catalog_objects = [Ent._catalog_entry_type() for Ent in entity_types]

    et = etree.parse(catalog_file)
    osc_root = et.getroot()
    catalog_element = osc_root.find("Catalog")
    catalog_name = catalog_element.attrib["name"]
    catalog = Catalog(catalog_name, relative_catalog_path)
    entries = {}
    for element in catalog_element.getchildren():
        entry = load_object(catalog, element, entity_types, catalog_objects)
        if entry is None:
            entry = Entity(CatalogEntry.from_xml(catalog, element))
        entries[entry.catalog_entry.catalog_entry] = entry
    return catalog, entries



from abc import ABC, abstractmethod
from typing import Any, List, Optional, Type

from scenario_gym.callback import StateCallback
from scenario_gym.state import State


class Metric(ABC):
    """
    Base class for a metric in scenario_gym.

    All metrics implement reset and step methods to update internal states during
    scenario rollout and the get_state method to return the metric value.

    The `required_callbacks` attribute can be set to a list of StateCallback
    subclasses. At reset the state will be checked to make sure each is present and
    the instance of each will be stored in `self.callbacks` in the same order as
    required callbacks.
    """

    name: Optional[str] = None
    required_callbacks: List[Type[StateCallback]] = []

    def __init__(self, name: Optional[str] = None):
        """
        Construct metric and set the name.

        Parameters
        ----------
        name : Optional[str]
            If not given then the name attribute will be used or
            the class name if that is None also.

        """
        if name is not None:
            self.name = name
        elif self.name is None:
            self.name = self.__class__.__name__
        self.callbacks: List[StateCallback] = []

    def reset(self, state: State) -> None:
        """Reset the metric at the start of a new scenario."""
        self.callbacks.clear()
        for CB in self.required_callbacks:
            cb = state.get_callback(CB)
            if cb is None:
                raise ValueError(
                    "Cannot run metric {} without callback {}.".format(
                        self.__class__.__name__,
                        CB.__name__,
                    )
                )
            self.callbacks.append(cb)
        self._reset(state)

    def step(self, state: State) -> None:
        """Update the metric after one timestep."""
        self._step(state)

    @abstractmethod
    def _reset(self, state: State) -> None:
        """Reset the metric at the start of a new scenario."""
        raise NotImplementedError

    @abstractmethod
    def _step(self, state: State) -> None:
        """Update the metric after one timestep."""
        raise NotImplementedError

    @abstractmethod
    def get_state(self) -> Any:
        """Return the current value of the metric."""
        raise NotImplementedError


def cache_metric(Met: Type[Metric]) -> Type[Metric]:
    """Wrap _step to cache the value whenver the state is terminal."""
    prev_step = Met._step
    Met.previous_value = None

    def new_step(self, state):
        prev_step(self, state)
        if state.is_done:
            self.previous_value = self.get_state()

    Met._step = new_step
    return Met


def cache_mean(Met: Type[Metric]) -> Type[Metric]:
    """Wrap _step to keep a running mean of the value."""

    def previous_value(self):
        val = self._previous_value
        self._previous_value = 0.0
        self._prev_count = 0
        return val

    prev_step = Met._step
    Met._previous_value = 0.0
    Met._prev_count = 0
    Met.previous_value = property(previous_value)

    def new_step(self, state):
        prev_step(self, state)
        if state.is_done:
            self._prev_count += 1
            self._previous_value += (
                self.get_state() - self._previous_value
            ) / self._prev_count

    Met._step = new_step
    return Met



from typing import Dict, List, Optional, TypeVar

import numpy as np
from scipy.interpolate import interp1d

from scenario_gym.trajectory import Trajectory
from scenario_gym.utils import ArrayLike

from .base import Entity

State = TypeVar("State")


class BatchReplayEntity:
    """
    A single object used to represent multiple entities.

    Will replay exact trajectories from OpenScenario files. Computation is
    vectorized for efficiency.
    """

    def __init__(self, timestep: Optional[float] = None):
        """Init the batch entity with no assigned entities."""
        self.entities: List[Entity] = []
        self.trajectories: List[Trajectory] = []
        self.timestep = timestep
        self.max_t = 0.0

    def step(self, state: State) -> Dict[Entity, ArrayLike]:
        """
        Take a single step in the gym.

        Returns the pose of each entity at the next timestamp.
        """
        new_poses = {}
        if len(self.entities) > 0:
            pos = self.fn(state.next_t)  # (m, num_ents)
            for e, p in zip(self.entities, pos):
                new_poses[e] = p
        return new_poses

    def add_entities(
        self,
        entities: List[Entity],
        trajs: List[Trajectory],
    ) -> None:
        """
        Add entities that are to be batched together.

        This will reset the entities in the scenario so all entities
        must be passed at once.

        Parameters
        ----------
        entities : List[Entity]
            The entities to be used.

        trajs : List[Trajectory]
            The trajectory for each entity.

        """
        self.entities.clear()
        self.trajectories.clear()
        self.max_t = 0.0
        if entities:
            self.entities.extend(entities)
            self.trajectories.extend(trajs)

            num_ents = len(self.entities)
            datas = []
            for t in self.trajectories:
                d = np.nan_to_num(t.data)
                if d.shape[0] == 1:
                    d = np.repeat(d, 2, axis=0)
                    d[-1, 0] += 1e-1  # to prevent nan
                datas.append(d)

            m = datas[0].shape[1] - 1
            ts = np.array(
                sorted(list(set([t for d in datas for t in d[:, 0]])))
            )  # (N,)
            self.max_t = ts[-1]

            interpd = []
            for d in datas:
                x = interp1d(
                    d[:, 0],
                    d[:, 1:].T,
                    bounds_error=False,
                    fill_value=(d[0, 1:], d[-1, 1:]),
                )(
                    ts
                ).T  # (N, m)
                interpd.append(x)

            X = np.concatenate(interpd, axis=1)  # (N, num_ents * m)
            if self.timestep:
                all_ts = np.arange(0.0, self.max_t, self.timestep)
                all_Xs = interp1d(
                    ts,
                    X.T,
                    bounds_error=False,
                    fill_value=(X[0], X[-1]),
                )(all_ts).T
                self.fn = lambda t: all_Xs[np.abs(all_ts - t).argmin()].reshape(
                    num_ents, m
                )
            else:
                interp = interp1d(
                    ts,
                    X.T,
                    bounds_error=False,
                    fill_value=(X[0], X[-1]),
                )
                self.fn = lambda t: interp(t).reshape(num_ents, m)



import numpy as np

from scenario_gym.controller import Controller
from scenario_gym.entity import Entity
from scenario_gym.pedestrian.action import PedestrianAction
from scenario_gym.state import State


class PedestrianController(Controller):
    """Pedestrian controller that applies a PedestrianAction."""

    def __init__(
        self,
        entity: Entity,
        max_speed: float = 5.0,
    ):
        """Init the pedestrian controller."""
        super().__init__(entity)
        self.max_speed = max_speed

    def _reset(self, state: State) -> None:
        """Reset the controller."""
        self.speed = 0.0

    def _step(self, state: State, action: PedestrianAction) -> np.ndarray:
        """
        Produce the new pose.

        Parameters
        ----------
        state : State
            Global state.

        action : PedestrianAction
            The action. Contains speed and heading angle.

        """
        pose = state.poses[self.entity].copy()
        self.speed = np.clip(action.speed, -self.max_speed, self.max_speed)
        pose[[0, 1]] += (
            self.speed
            * state.dt
            * np.array([np.cos(action.heading), np.sin(action.heading)])
        )
        pose[3] = action.heading
        return pose



from dataclasses import dataclass
from typing import Any, Dict, Optional

from lxml.etree import Element

from scenario_gym.catalog_entry import (
    ArgsKwargs,
    Catalog,
    CatalogEntry,
    CatalogObject,
)
from scenario_gym.entity.base import Entity
from scenario_gym.trajectory import Trajectory


@dataclass
class Axle(CatalogObject):
    """A front or rear axle of a vehicle."""

    max_steering: float
    wheel_diameter: float
    track_width: float
    position_x: float
    position_z: float

    @classmethod
    def load_data_from_xml(cls, catalog_name: str, element: Element) -> ArgsKwargs:
        """Load the bounding box data form an xml element."""
        return (
            float(element.attrib["maxSteering"]),
            float(element.attrib["wheelDiameter"]),
            float(element.attrib["trackWidth"]),
            float(element.attrib["positionX"]),
            float(element.attrib["positionZ"]),
        ), {}

    def to_dict(self) -> None:
        """Write the vehicle catalog entry to a dictionary."""
        return {
            "max_steering": self.max_steering,
            "wheel_diameter": self.wheel_diameter,
            "track_width": self.track_width,
            "position_x": self.position_x,
            "position_z": self.position_z,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Load the vehicle catalog entry from a dictionary."""
        return cls(
            data["max_steering"],
            data["wheel_diameter"],
            data["track_width"],
            data["position_x"],
            data["position_z"],
        )


@dataclass
class VehicleCatalogEntry(CatalogEntry):
    """Catalog entry for a vehicle."""

    mass: Optional[float]
    max_speed: Optional[float]
    max_deceleration: Optional[float]
    max_acceleration: Optional[float]
    front_axle: Optional[Axle]
    rear_axle: Optional[Axle]

    xosc_names = ["Vehicle"]

    @classmethod
    def load_data_from_xml(cls, catalog_name: str, element: Element) -> ArgsKwargs:
        """Load the vehicle from an xml element."""
        base_args, _ = super().load_data_from_xml(catalog_name, element)
        performance = element.find("Performance")
        front_axle = element.find("Axles/FrontAxle")
        rear_axle = element.find("Axles/RearAxle")
        mass = float(element.attrib["mass"]) if "mass" in element.attrib else None
        if performance is not None:
            max_speed = float(performance.attrib["maxSpeed"])
            max_dec = float(performance.attrib["maxDeceleration"])
            max_acc = float(performance.attrib["maxAcceleration"])
        else:
            max_speed = max_dec = max_acc = None
        front_axle = (
            Axle.from_xml(catalog_name, front_axle)
            if front_axle is not None
            else None
        )
        rear_axle = (
            Axle.from_xml(catalog_name, rear_axle)
            if rear_axle is not None
            else None
        )
        veh_args = (
            mass,
            max_dec,
            max_acc,
            max_speed,
            front_axle,
            rear_axle,
        )
        return base_args + veh_args, {}

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Load the vehicle from a dictionary."""
        return cls(
            Catalog(data["catalog"]["catalog_name"], data["catalog"]["rel_path"]),
            data["catalog_entry"],
            data["catalog_category"],
            data["catalog_type"],
            data["bounding_box"],
            data.get("properties", {}),
            data.get("files", []),
            data.get("mass"),
            data.get("max_speed"),
            data.get("max_deceleration"),
            data.get("max_acceleration"),
            Axle.from_dict(data["front_axle"]) if "front_axle" in data else None,
            Axle.from_dict(data["rear_axle"]) if "rear_axle" in data else None,
        )

    def to_dict(self) -> Dict[str, Any]:
        """Write the scenario to a dictionary."""
        data = super().to_dict()
        data.update(
            {
                "mass": self.mass,
                "max_speed": self.max_speed,
                "max_deceleration": self.max_deceleration,
                "max_acceleration": self.max_acceleration,
                "front_axle": self.front_axle.to_dict()
                if self.front_axle is not None
                else None,
                "rear_axle": self.rear_axle.to_dict()
                if self.rear_axle is not None
                else None,
            }
        )
        return data


class Vehicle(Entity):
    """Class for vehicles."""

    def __init__(
        self,
        catalog_entry: VehicleCatalogEntry,
        trajectory: Optional[Trajectory] = None,
        ref: Optional[str] = None,
    ):
        super().__init__(catalog_entry, trajectory=trajectory, ref=ref)



from __future__ import annotations

from copy import copy
from typing import Callable, List, Optional, Tuple

import numpy as np
from scipy.interpolate import interp1d

from scenario_gym.utils import ArrayLike, NDArray, cached_property


class Trajectory:
    """
    A Scenario Gym representation of a trajectory.

    Note that trajectories consist of immutable arrays. To modify a trajectory
    one must copy the data and init a new one:
    ```
    new_data = trajectory.data.copy()
    # apply changes
    new_t = Trajectory(new_data)
    ```
    """

    _fields = ("t", "x", "y", "z", "h", "p", "r")
    t: Optional[NDArray] = None
    x: Optional[NDArray] = None
    y: Optional[NDArray] = None
    z: Optional[NDArray] = None
    h: Optional[NDArray] = None
    p: Optional[NDArray] = None
    r: Optional[NDArray] = None

    def __init__(self, data: NDArray, fields: Tuple[str] = _fields):
        """
        Trajectory constructor.

        Parameters
        ----------
        data : np.ndarray
            The trajectory data as a numpy array of (num_points, num_fields).
            By default the columns should be t, x, y, z, h, p, r otherwise the
            fields argument should be passed.

        fields : List[str]
            The field names for each column of data. Must contain t, x and y and
            must be a subset of _fields.

        """
        if not all(f in fields for f in ("t", "x", "y")):
            raise ValueError("Trajectory cannot be created with t, x and y values.")
        if data.ndim != 2 or data.shape[1] != len(fields):
            raise ValueError(
                f"Invalid shape: {data.shape}. Expected: (N, {len(fields)}). Either"
                " pass `fields` to specify the columns given or ensure that columns"
                f" for all of {self._fields} are provided."
            )
        perm = [fields.index(f) for f in self._fields if f in fields]
        data = data[:, perm]
        data = np.unique(np.where(np.isfinite(data), data, np.inf), axis=0)
        n = data.shape[0]

        _data: List[NDArray] = []
        for f in self._fields:
            d = data[:, perm[fields.index(f)]] if f in fields else np.zeros(n)
            if f == "h":
                d = _resolve_heading(d)
            if f not in fields or (f in fields and np.isfinite(d).sum() != n):
                if f == "h" and n == 1:
                    d = np.zeros(1)
                elif f == "h" and n > 1:
                    t = _data[0]
                    fn = interp1d(
                        t,
                        np.array(_data[1:3]).T,
                        axis=0,
                        fill_value="extrapolate",
                    )
                    d = np.arctan2(*np.flip(fn(t + 1e-2) - fn(t - 1e-2), axis=1).T)
                    d = _resolve_heading(d)
                elif f in ("z", "p", "r"):
                    d = np.zeros(n)
                else:
                    raise ValueError(
                        f"Invalid values found for {f}. Values required for xyt."
                    )
            _data.append(d)
            setattr(self, f, d)

        # we will make the data readonly
        self.data = np.array(_data).T.copy()
        self.data.flags.writeable = False

        self._interpolated: Optional[Callable[[ArrayLike], NDArray]] = None
        self._interpolated_s: Optional[Callable[[ArrayLike], NDArray]] = None
        self._grad_fn = None

    def __len__(self) -> int:
        """Return the number of points in the trajectory."""
        return len(self.data)

    def __getitem__(self, idx: int) -> NDArray:
        """Get the idx'th point in the trajectory."""
        return self.data[idx]

    @cached_property
    def min_t(self) -> float:
        """Return the first timestamp of the trajectory."""
        return self.t.min()

    @cached_property
    def max_t(self) -> float:
        """Return the final timestamp of the trajectory."""
        return self.t.max()

    @cached_property
    def s(self) -> NDArray:
        """Return the distance travelled at each point."""
        ds = np.linalg.norm(np.diff(self.data[:, [1, 2]], axis=0), axis=1).cumsum()
        return np.hstack([[0.0], ds])

    @cached_property
    def arclength(self) -> float:
        """Total distance travelled."""
        return self.s[-1]

    def position_at_t(self, t: float) -> NDArray:
        """
        Compute the position of the entity at time t.

        Parameters
        ----------
        t : float
            The time at which the position is returned. Linearly interpolates
            the trajectory control points to find the position.

        Returns
        -------
        np.ndarray
            The position as a numpy array.

        """
        if self._interpolated is None:
            data = self.data
            if data.shape[0] == 1:
                data = np.repeat(data, 2, axis=0)
                data[-1, 0] += 1e-3
            self._interpolated = interp1d(
                data[:, 0],
                data[:, 1:],
                bounds_error=False,
                fill_value=(data[0, 1:], data[-1, 1:]),
                axis=0,
            )
        return self._interpolated(t)

    def position_at_s(self, s: float) -> NDArray:
        """
        Compute the position of the entity at distance travelled s.

        Parameters
        ----------
        s : float
            The arclength at which the position is returned. Linearly
            interpolates the trajectory control points to find the position.

        Returns
        -------
        np.ndarray
            The position as a numpy array.

        """
        if self._interpolated_s is None:
            data = self.data
            if data.shape[0] == 1:
                data = np.repeat(data, 2, axis=0)
                data[-1, 0] += 1e-3
            self._interpolated_s = interp1d(
                self.s,
                data,
                bounds_error=False,
                fill_value=(data[0, :], data[-1, :]),
                axis=0,
            )
        return self._interpolated_s(s)

    def is_stationary(self) -> bool:
        """Return True if the trajectory is stationary."""
        return is_stationary(self.data)

    def __copy__(self) -> Trajectory:
        """Create a copy of the trajectory."""
        return self.__class__(self.data.copy())

    def copy(self) -> Trajectory:
        """Create a copy of the trajectory."""
        return copy(self)

    def translate(self, x: np.ndarray) -> Trajectory:
        """
        Create a new trajectory by translating the current by x.

        Parameters
        ----------
        x : np.ndarray
            Translation quantity. Must broadcast to the data so must be
            a matrix, vector or scalar.

        Returns
        -------
        Trajectory
            The translated trajectory.

        """
        if x.ndim == 1:
            x = x[None, :]
        return self.__class__(self.data + x)

    def rotate(self, h: float) -> Trajectory:
        """
        Create a new trajectory by rotating the current by h around O.

        Parameters
        ----------
        h : float
            The angle of rotation (about the origin).

        Returns
        -------
        Trajectory
            The rotated trajectory.

        """
        new_data = self.data.copy()
        xy = new_data[None, 0, [1, 2]]
        new_data[:, [1, 2]] = (new_data[:, [1, 2]] - xy).dot(
            np.array(
                [
                    [np.cos(h), np.sin(h)],
                    [-np.sin(h), np.cos(h)],
                ]
            )
        ) + xy
        new_data[:, 4] = (new_data[:, 4] + h) % (2.0 * np.pi)
        return self.__class__(new_data)

    def subsample(
        self,
        points_per_s: Optional[float] = None,
        points_per_t: Optional[float] = None,
        curvature: bool = False,
        **kwargs,
    ) -> Trajectory:
        """
        Create a new trajectory with a given frequency of control points.

        The control points can either be equally spaced across time or across arc by
        passing the keyword arguments `points_per_t` or `points_per_s` respectively.
        Exactly one of these keywords must be passed.

        Parameters
        ----------
        points_per_s : Optional[float]
            Number of control points per unit of arc.

        points_per_t : Optional[float]
            Number of control points per unit of time.

        curvature: bool
            If given will given curvature sampling to subsample the trajectory.

        """
        if (points_per_s is None) == (points_per_t is None):
            raise ValueError(
                "Exactly one of `points_per_s` or `points_per_t` must be supplied."
            )
        if curvature:
            return self.curvature_subsample(
                points_per_s=points_per_s,
                points_per_t=points_per_t,
                **kwargs,
            )
        if points_per_t:
            n = int(np.ceil((self.max_t - self.min_t) * points_per_t))
            ts = np.linspace(self.min_t, self.max_t, n)
            data = self.position_at_t(ts)
            return self.__class__(np.concatenate([ts[:, None], data], axis=1))

        n = int(np.ceil(self.arclength * points_per_s))
        ss = np.linspace(0, self.arclength, n)
        data = self.position_at_s(ss)
        return self.__class__(data)

    def curvature_subsample(
        self,
        points_per_s: Optional[float] = None,
        points_per_t: Optional[float] = None,
        eps: float = 1e-3,
        weight: float = 5.0,
    ) -> np.ndarray:
        """
        Subsample by sampling points arround high curvature areas.

        Parameters
        ----------
        points_per_s : Optional[float]
            Number of control points per unit of arc.

        points_per_t : Optional[float]
            Number of control points per unit of time.

        eps : float
            Epsilon parameter for computing gradients of the trajectory.

        weight : float
            Temperature for sampling distribution. Higher values give points more
            densley sampled around high curvature areas.

        """
        if points_per_s is not None:
            n = int(np.maximum(1, points_per_s * self.arclength))
        elif points_per_t is not None:
            n = int(np.maximum(1, points_per_t * self.max_t))
        else:
            raise ValueError(
                "Exactly one of `points_per_s` or `points_per_t` must be supplied."
            )
        s = self.s
        if self._grad_fn is None:
            fn = self.position_at_s
            grads = (fn(s + eps)[:, [1, 2]] - fn(s - eps)[:, [1, 2]]) / (2 * eps)
            self._grad_fn = interp1d(s, grads, axis=0, fill_value="extrapolate")
        grad_fn = self._grad_fn
        second_grad = (grad_fn(s[1:-1] + eps) - grad_fn(s[1:-1] - eps)) / (2 * eps)
        curv = np.linalg.norm(second_grad, axis=1)
        dist = np.exp(weight * curv) / np.exp(weight * curv).sum()
        num_points = int(np.clip(n - 2, 1, dist.shape[0]))
        idxs = np.random.choice(
            dist.shape[0],
            size=(num_points,),
            replace=False,
            p=dist,
        )
        s_vals = s[np.hstack([[0], 1 + np.sort(idxs), [s.shape[0] - 1]])]
        return self.__class__(fn(s_vals))

    def to_json(self) -> List[List[float]]:
        """Write the trajectory to a jsonable list."""
        return self.data.tolist()


def _resolve_heading(h: NDArray) -> NDArray:
    """Update heading so that there are no large jumps."""
    deltas = np.diff(h) % (2 * np.pi)
    deltas = np.where(deltas > np.pi, deltas - 2 * np.pi, deltas)
    return np.hstack([h[0], deltas]).cumsum()


def is_stationary(data: np.ndarray) -> bool:
    """
    Check if an entity is stationary for the entire scenario.

    Any nan values are replaced with 0s.
    """
    return (
        len(
            np.unique(
                np.where(
                    np.isnan(data[:, 1:]),
                    0.0,
                    data[:, 1:],
                ),
                axis=0,
            )
        )
        <= 1
    )



from dataclasses import dataclass
from typing import List, Tuple

from shapely.geometry import MultiPolygon

from scenario_gym.entity import Entity
from scenario_gym.observation import SingleEntityObservation
from scenario_gym.utils import NDArray


@dataclass
class PedestrianObservation(SingleEntityObservation):
    """
    An observation for a pedestrian agent.

    Contains the objects and road elements detected within a given radius.

    Todo: Instead of returning complete state, return observation with only entities
    and pedestrian road info within a given radius by adapting RasterizedMapSensor
    or ObjectDetectionSensor.
    """

    head_rot_angle: float
    near_peds: List[Tuple[Entity, NDArray, NDArray]]
    walkable_surface: MultiPolygon
    impenetrable_surface: MultiPolygon



"""Parameters for pedestrian behavior models."""

# General parameters for all models
general_params = {"speed": 0.4, "max_speed_factor": 1.3}

# Random walk parameters
random_walk = {
    "bias_lon": 0.1,
    "bias_lat": 0.05,
    "std_lon": 0.2,
    "std_lat": 0.1,
}

social_force = {
    "distance_threshold": 3,
    "sight_weight": 0.5,
    "sight_angle": 200,
    "relaxation_time": 1.5,
    "ped_repulse_V": 2.1,
    "ped_repulse_sigma": 0.3,
    "ped_attract_C": 0.2,
    "boundary_repulse_U": 10,
    "boundary_repulse_R": 0.2,
}

# Add models to be used here
models_params = {"random_walk": random_walk, "social_force": social_force}

params = {"general": general_params, "models": models_params}



from enum import Enum
from typing import Any, Dict, List, Optional, Union

import numpy as np
from shapely.geometry import LineString, Polygon

from scenario_gym.utils import ArgsKwargs

from .base import RoadGeometry, RoadLike


class LaneType(Enum):
    """Enumerates OpenDrive standard lane types."""

    none = 0
    driving = 1
    biking = 3
    border = 4
    connectingRamp = 5
    curb = 6
    entry = 7
    exit = 8
    median = 9
    offRamp = 10
    onRamp = 11
    parking = 12
    restricted = 13
    sidewalk = 14
    shoulder = 15
    stop = 16


class Lane(RoadLike):
    """
    A lane object.

    The lane has a center and a boundary and holds information of its successor
    and predeceessor lanes.
    """

    walkable = False

    @classmethod
    def load_data_from_dict(cls, l: Dict[str, Any]) -> ArgsKwargs:
        """Create from dictionary."""
        args, kwargs = super().load_data_from_dict(l)
        typ = l["type"] if "type" in l else "driving"
        lane_type = LaneType[typ if typ in LaneType.__members__ else "driving"]
        return (
            *args,
            l["successors"] if "successors" in l else [],
            l["predecessors"] if "predecessors" in l else [],
            lane_type,
        ), kwargs

    def __init__(
        self,
        id: str,
        boundary: Polygon,
        center: LineString,
        successors: List[str],
        predecessors: List[str],
        _type: Union[str, LaneType],
        elevation: Optional[np.ndarray] = None,
    ):
        super().__init__(id, boundary, center, elevation=elevation)
        self.successors = successors
        self.predecessors = predecessors
        if isinstance(_type, str):
            if _type not in LaneType.__members__:
                raise ValueError(
                    f"{self.type} is not a valid lane type. "
                    "Check objects.py to see all valid lane types.",
                )
            else:
                _type = LaneType[_type]
        self._type = _type

    @property
    def type(self) -> LaneType:
        """Get the type of the lane."""
        return self._type

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary of the objects data."""
        data = super().to_dict()
        data.update(
            {
                "successors": self.successors,
                "predecessors": self.predecessors,
            }
        )
        if self.type is not None:
            data["type"] = self.type.name
        return data


class Road(RoadLike):
    """
    A road object.

    The road has a center and a boundary and lanes.
    """

    walkable = False

    @classmethod
    def load_data_from_dict(cls, r: Dict[str, Any]) -> ArgsKwargs:
        """Create from dictionary."""
        args, kwargs = super().load_data_from_dict(r)
        lanes = [Lane.from_dict(l) for l in r["lanes" if "lanes" in r else "Lanes"]]
        return (*args, lanes), kwargs

    def __init__(
        self,
        id: str,
        boundary: Polygon,
        center: LineString,
        lanes: List[Lane],
        elevation: Optional[np.ndarray] = None,
    ):
        super().__init__(id, boundary, center, elevation=elevation)
        self.lanes = lanes

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary of the objects data."""
        data = super().to_dict()
        data["lanes"] = [l.to_dict() for l in self.lanes]
        return data


class Intersection(RoadGeometry):
    """
    An intersection object.

    The intersection has a boundary, connecting lanes and the ids of
    the roads it connects.
    """

    driveable = True
    walkable = False

    @classmethod
    def load_data_from_dict(cls, i: Dict[str, Any]) -> ArgsKwargs:
        """Create from dictionary."""
        args, kwargs = super().load_data_from_dict(i)
        lanes = [Lane.from_dict(l) for l in i["lanes" if "lanes" in i else "Lanes"]]
        return (*args, lanes, i["connecting_roads"]), kwargs

    def __init__(
        self,
        id: str,
        boundary: Polygon,
        lanes: List[Lane],
        connecting_roads: List[str],
        elevation: Optional[np.ndarray] = None,
    ):
        super().__init__(id, boundary, elevation=elevation)
        self.lanes = lanes
        self.connecting_roads = connecting_roads

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary of the objects data."""
        data = super().to_dict()
        data.update(
            {
                "lanes": [l.to_dict() for l in self.lanes],
                "connecting_roads": self.connecting_roads,
            }
        )
        return data


class Pavement(RoadLike):
    """
    A pavement object.

    The pavement has a boundary and a center.
    """

    driveable = False


class Crossing(RoadLike):
    """
    A crossing object.

    The crossing has a boundary and center and ids of pavements it connects.
    """

    driveable = False

    @classmethod
    def load_data_from_dict(cls, c: Dict[str, Any]) -> ArgsKwargs:
        """Create from dictionary."""
        args, kwargs = super().load_data_from_dict(c)
        return (
            *args,
            c["pavements" if "pavements" in c else "Pavements"],
        ), kwargs

    def __init__(
        self,
        id: str,
        boundary: Polygon,
        center: LineString,
        pavements: List[str],
        elevation: Optional[np.ndarray] = None,
    ):
        super().__init__(id, boundary, center, elevation=elevation)
        self.pavements = pavements

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary of the objects data."""
        data = super().to_dict()
        data["pavements"] = self.pavements
        return data


class Building(RoadGeometry):
    """
    A geometry describing the area of a building.

    These are modelled as solid blocks that cannot be
    entered by vehicles or pedestrians.
    """

    driveable = False
    impenetrable = True



import os
from typing import List, Optional
from xml.etree import ElementTree as ET

import numpy as np
from scenariogeneration import xosc

from scenario_gym.entity import Entity
from scenario_gym.scenario import Scenario
from scenario_gym.trajectory import is_stationary


def write_scenario(
    scenario: Scenario,
    filepath: str,
    base_road_network_path: str = "../Road_Networks",
    osc_minor_version: int = 0,
) -> None:
    """
    Write a scenario to an OpenScenario file.

    Parameters
    ----------
    scenario : Scenario
        The scenario object.

    filepath : str
        The desired filepath.

    base_road_network_path : str
        Base path to the road networks.

    osc_minor_version : int
        The OpenScenario minor version.

    """
    name = (
        scenario.name
        if scenario.name is not None
        else (os.path.splitext(os.path.basename(scenario.path)[-1])[0])
    )

    rn_name = scenario.road_network.path.split("/")[-1].split(".")[0]
    scenegraph = os.path.join(base_road_network_path, f"{rn_name}.json")
    rn = xosc.RoadNetwork("", scenegraph)

    entities = xosc.Entities()
    catalog = xosc.Catalog()
    for e in scenario.entities:
        ce = e.catalog_entry
        if ce.catalog_type not in catalog.catalogs:
            catalog_dir = scenario.catalog_locations[ce.catalog_name]
            catalog.add_catalog(f"{ce.catalog_type}Catalog", catalog_dir)
        catalog_ref = xosc.CatalogReference(ce.catalog_name, ce.catalog_entry)
        entities.add_scenario_object(e.ref, catalog_ref)

    init = xosc.Init()
    for e in scenario.entities:
        if is_stationary(e.trajectory.data[:, 1:]):
            pose = e.trajectory.data[0, 1:]
            if not np.isfinite(pose[3]):
                raise ValueError(f"Heading must be finite but is {pose[2]}.")
            action = xosc.TeleportAction(
                xosc.WorldPosition(
                    *(float(p) if np.isfinite(p) else None for p in pose)
                )
            )
            init.add_init_action(e.ref, action)

    act = xosc.Act(
        scenario.path.replace(".xosc", ""),
        get_simulation_time_trigger(0),
    )
    maneuver_groups = []
    for idx, e in enumerate(scenario.entities):
        m_group = get_maneuver_group(
            e, osc_minor_version=osc_minor_version, check_stationary=(idx > 0)
        )
        if m_group:
            maneuver_groups.append(m_group)
            act.add_maneuver_group(m_group)

    story = xosc.Story(name)
    story.add_act(act)
    sb = xosc.StoryBoard(init)
    sb.add_story(story)

    desc = (
        f"Scenario {name} recorded in the dRISK Scenario Gym subject to the dRISK "
        "License Agreement (https://drisk.ai/license/)."
    )
    s = xosc.Scenario(
        desc,
        "∂RISK",
        xosc.ParameterDeclarations(),
        entities=entities,
        storyboard=sb,
        roadnetwork=rn,
        catalog=catalog,
        osc_minor_version=osc_minor_version,
    )
    element = ET.Element("OpenSCENARIO")
    element.extend(
        (
            s.header.get_element(),
            s.parameters.get_element(),
            s.catalog.get_element(),
            s.roadnetwork.get_element(),
            s.entities.get_element(),
            s.storyboard.get_element(),
        )
    )
    s.write_xml(filepath)


def get_simulation_time_trigger(t: float, delay: float = 0.0) -> xosc.ValueTrigger:
    """Get a simulation time trigger."""
    return xosc.ValueTrigger(
        "startSimTrigger",
        delay=delay,
        conditionedge=xosc.ConditionEdge.rising,
        valuecondition=xosc.SimulationTimeCondition(
            value=t, rule=xosc.Rule.greaterThan
        ),
    )


def get_follow_trajectory_event(
    e: Entity,
    osc_minor_version: int = 0,
    check_stationary: bool = True,
) -> Optional[xosc.Event]:
    """Get a follow trajectory event for an entity."""
    if check_stationary and is_stationary(e.trajectory.data):
        return None
    ts, poses = e.trajectory.t, e.trajectory.data[:, 1:]
    positions = [
        xosc.WorldPosition(*(float(p) if np.isfinite(p) else None for p in pose))
        for pose in poses
    ]
    polyline = xosc.Polyline(ts.tolist(), positions)
    traj = xosc.Trajectory(f"{e.ref}_trajectory", False)
    traj.add_shape(polyline)
    follow_trajectory_action = xosc.FollowTrajectoryAction(
        traj,
        following_mode=xosc.FollowMode.position,
        reference_domain=xosc.ReferenceContext.absolute,
        scale=1,
        offset=0,
    )
    follow_trajectory_action.version_minor = osc_minor_version
    follow_trajectory_event = xosc.Event(
        f"{e.ref}_follow_trajectory_event",
        xosc.Priority.overwrite,
    )
    follow_trajectory_event.add_action(
        "follow_trajectory_action",
        follow_trajectory_action,
    )
    follow_trajectory_event.add_trigger(get_simulation_time_trigger(0))
    return follow_trajectory_event


def get_events(
    e: Entity, osc_minor_version: int = 0, check_stationary: bool = True
) -> List[xosc.Event]:
    """Get events for the given entity."""
    events = []
    follow_trajectory_event = get_follow_trajectory_event(
        e,
        osc_minor_version=osc_minor_version,
        check_stationary=check_stationary,
    )
    if follow_trajectory_event:
        events.append(follow_trajectory_event)
    # NOTE: ignoring non-trajectory actions for now
    return events


def get_maneuver(
    e: Entity, osc_minor_version: int = 0, check_stationary: bool = True
) -> Optional[xosc.Maneuver]:
    """Get maneuvers for the given entity."""
    events = get_events(
        e, osc_minor_version=osc_minor_version, check_stationary=check_stationary
    )
    if events:
        maneuver = xosc.Maneuver(f"{e.ref}_maneuver")
        for event in events:
            maneuver.add_event(event)
        return maneuver


def get_maneuver_group(
    e: Entity, osc_minor_version: int = 0, check_stationary: bool = True
) -> Optional[xosc.ManeuverGroup]:
    """Get the maneuver group for the given entity."""
    maneuver = get_maneuver(
        e, osc_minor_version=osc_minor_version, check_stationary=check_stationary
    )
    if maneuver:
        mangrp = xosc.ManeuverGroup(f"{e.ref}_maneuver_group")
        mangrp.add_actor(e.ref)
        mangrp.add_maneuver(maneuver)
        return mangrp



from typing import Tuple, Union

import numpy as np
from shapely.geometry import MultiPolygon, Point, Polygon
from shapely.ops import nearest_points

from scenario_gym.agent import Agent
from scenario_gym.entity import Entity
from scenario_gym.pedestrian.behaviour import PedestrianBehaviour
from scenario_gym.pedestrian.observation import PedestrianObservation
from scenario_gym.pedestrian.random_walk import RandomWalkParameters
from scenario_gym.utils import NDArray
from scenario_gym.viewer import rotate_coords


class SocialForceParameters(RandomWalkParameters):
    """Parameters for the social force model."""

    distance_threshold = 3
    sight_weight = 0.5
    sight_weight_use = True
    sight_angle = 200
    relaxation_time = 1.5
    ped_repulse_V = 1.0
    ped_repulse_sigma = 1.0
    ped_attract_C = 0.0
    boundary_repulse_U = 10.0
    boundary_repulse_R = 0.2
    imp_boundary_repulse_U = 2.0
    imp_boundary_repulse_R = 0.1


class SocialForce(PedestrianBehaviour):
    """Social force model."""

    def __init__(self, params: SocialForceParameters):
        """Init the behaviour model."""
        super().__init__(params)
        self.bias_lon = params.bias_lon
        self.bias_lat = params.bias_lat
        self.std_lon = params.std_lon
        self.std_lat = params.std_lat

    def _step(self, observation: PedestrianObservation, agent: Agent) -> Tuple:
        """Return the new speed and heading using the social force model."""
        # Start with attraction force to goal point
        force_sum = self._force_to_goal(
            observation,
            agent.route[agent.goal_idx],
            agent.speed_desired,
        )

        for (
            pedestrian,
            pose,
            vel,
        ) in observation.near_peds:  # Forces from other pedestrians
            # Vector of agent's sight (velocity angle + head angle)
            view_dir_vector = rotate_coords(vel[[0, 1]], observation.head_rot_angle)
            view_dir_unit_vector = view_dir_vector / (
                np.linalg.norm(view_dir_vector) + 0.0000000001
            )

            force_repulsion = self._force_pedestrian_repulsion(
                observation, (pedestrian, pose, vel)
            )
            force_attraction = self._force_pedestrian_attraction(
                observation, (pedestrian, pose, vel)
            )
            if self.params.sight_weight_use:
                force_sum += (
                    self._sight_weight(force_repulsion, view_dir_unit_vector)
                    * force_repulsion
                )
                force_sum += (
                    self._sight_weight(force_attraction, view_dir_unit_vector)
                    * force_attraction
                )
            else:
                force_sum += force_attraction
                force_sum += force_repulsion

        # get current position
        point = Point(*observation.pose[:2])

        # Force from closest walkable boundary
        if observation.walkable_surface.area > 0:
            if observation.walkable_surface.contains(point):
                force_sum += self._force_boundary(
                    observation,
                    observation.walkable_surface,
                    self.params.boundary_repulse_R,
                    self.params.boundary_repulse_U,
                )

        # Force from immovable boundary
        if observation.impenetrable_surface.area > 0:
            sign = 1 - 2 * observation.impenetrable_surface.contains(point)
            force_sum += sign * self._force_boundary(
                observation,
                observation.impenetrable_surface,
                self.params.imp_boundary_repulse_R,
                self.params.imp_boundary_repulse_U,
            )

        # Random fluctuations
        speed_rand = np.random.normal(self.bias_lon, self.std_lon)
        heading_rand = np.random.normal(self.bias_lat, self.std_lat)

        speed = min(
            np.linalg.norm(force_sum) + speed_rand,
            agent.speed_desired * self.max_speed_factor,
        )  # Limit to max speed
        heading = np.arctan2(force_sum[1], force_sum[0]) + heading_rand
        agent.force = force_sum

        return speed, heading

    def _force_to_goal(
        self,
        obs: PedestrianObservation,
        goal_point: NDArray,
        speed_desired: float,
    ) -> np.ndarray:
        """Compute the attraction force from the goal."""
        agent_pos = obs.pose[[0, 1]]
        agent_vel = obs.velocity[[0, 1]]
        dir_vector = goal_point - agent_pos
        dir_vector_norm = np.linalg.norm(dir_vector)
        if dir_vector_norm == 0:
            dir_vector_norm += 0.000000001
        unit_dir_vector = dir_vector / dir_vector_norm
        force_vector = (
            1
            / self.params.relaxation_time
            * (speed_desired * unit_dir_vector - agent_vel)
        )
        return force_vector

    def _force_pedestrian_repulsion(
        self,
        obs: PedestrianObservation,
        other_pedestrian: Tuple[Entity, NDArray, NDArray],
    ) -> NDArray:
        """Compute the repulsion force from other pedestrians."""
        agent_pos = obs.pose[[0, 1]]
        other_ped, other_pose, other_v = other_pedestrian
        other_pos = other_pose[[0, 1]]
        other_dir = other_v[[0, 1]]

        # Vector to other agent
        r_ao = agent_pos - other_pos
        r_ao_norm = np.linalg.norm(r_ao)

        # Auxiliary calculations
        v_vel_magnitude = np.linalg.norm(other_dir) + 0.0000000001
        unit_other_dir = other_dir / v_vel_magnitude
        other_step = v_vel_magnitude * (obs.next_t - obs.t)
        r_ao_other = r_ao - other_step * unit_other_dir
        r_ao_other_norm = np.linalg.norm(r_ao_other) + 0.0000000001

        # Ellipse semi-minor axis b
        b = (1 / 2) * np.sqrt((r_ao_norm + r_ao_other_norm) ** 2 - other_step**2)
        db = (
            (1 / 4)
            * (1 / b)
            * (r_ao_norm + r_ao_other_norm)
            * (r_ao / r_ao_norm + r_ao_other / r_ao_other_norm)
        )
        force_vector = (
            self.params.ped_repulse_V
            / self.params.ped_repulse_sigma
            * np.exp(-b / self.params.ped_repulse_sigma)
            * db
        )  # gradient
        return force_vector

    def _force_pedestrian_attraction(
        self,
        obs: PedestrianObservation,
        other_pedestrian: Tuple[Entity, NDArray, NDArray],
    ) -> NDArray:
        """Compute the attraction force from other pedestrians."""
        agent_pos = obs.pose[[0, 1]]
        other_pos = other_pedestrian[1][[0, 1]]
        # Vector to other agent
        r_ao = agent_pos - other_pos
        return 2 * self.params.ped_attract_C * r_ao  # gradient

    def _force_boundary(
        self,
        obs: PedestrianObservation,
        object: Union[Polygon, MultiPolygon],
        param_r: float,
        param_u: float,
    ) -> NDArray:
        """
        Compute the force from the boundary of an object.

        Can be an attractive or a repulsive force.
        """
        agent_pos = obs.pose[[0, 1]]
        agent_point = Point(*agent_pos)
        closest_point, _ = nearest_points(object, agent_point)
        closest_pos = np.array(closest_point.xy).squeeze()
        r_aB = agent_pos - closest_pos
        r_aB_norm = np.linalg.norm(r_aB)
        r_aB_unit = r_aB / (r_aB_norm + 0.0000000001)
        return (
            param_u / param_r * r_aB_unit * np.exp(-r_aB_norm / param_r)
        )  # gradient

    def _sight_weight(
        self, force_vector: np.ndarray, view_dir_unit_vector: np.ndarray
    ) -> float:
        """Compute the weight force depending on angle of sight."""
        dot_dir = np.dot(view_dir_unit_vector, force_vector) / (
            np.linalg.norm(force_vector) + 0.0000000001
        )
        if dot_dir >= np.cos(self.params.sight_angle / 2 * np.pi / 180):
            return 1.0
        return self.params.sight_weight



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, TypeVar

from scenario_gym.entity import Entity

State = TypeVar("State")


class ScenarioAction(ABC):
    """
    Base class for scenario actions.

    Actions are applied at the first timestamp with time greater than or equal to
    the action time. They are applied with the _apply method which must be
    implemented.

    """

    def __init__(
        self,
        t: float,
        action_class: str,
        entity_ref: str,
        action_variables: Dict[str, Any],
    ):
        """
        Create the action.

        Parameters
        ----------
        t : float
            The time at which the action should be applied.

        action_class : str
            The name of the action class.

        entity_ref : str
            Reference of the entity to which the action applies.

        action_variables :  Dict[str, Any]
            Dictionary of action variables.

        """
        self.t = t
        self.action_class = action_class
        self.entity_ref = entity_ref
        self.action_variables = action_variables

    def apply(self, state: State, entity: Optional[Entity]) -> None:
        """Apply the action to the environment state."""
        self._apply(state, entity)
        self._applied = True

    @abstractmethod
    def _apply(self, state: State, entity: Optional[Entity]) -> None:
        """Apply the action to the environment state."""
        raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        """Write the action to a dictionary."""
        return {
            "t": self.t,
            "action_class": self.action_class,
            "entity_ref": self.entity_ref,
            "action_variables": self.action_variables,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Load the action from a dictionary."""
        return cls(
            data["t"],
            data["action_class"],
            data["entity_ref"],
            data["action_variables"],
        )


class UpdateStateVariableAction(ScenarioAction):
    """Action that sets state variables for the entity."""

    def _apply(self, state: State, entity: Optional[Entity]) -> None:
        """Update the entity with action variables."""
        if entity is not None:
            if state.entity_state[entity] is None:
                state.entity_state[entity] = {}
            for k, v in self.action_variables.items():
                state.entity_state[entity][k] = v



import math

import numpy as np


def rotate_coords(
    X: np.ndarray,
    theta: float,
) -> np.ndarray:
    """Rotate xy coordinates in X by angle theta."""
    R = np.array(
        [
            [math.cos(theta), -math.sin(theta)],
            [math.sin(theta), math.cos(theta)],
        ]
    )
    return X.dot(R.T)


def vec2pix(
    X: np.ndarray, mag: int = 20, h: float = 100.0, w: float = 100.0
) -> np.ndarray:
    """Convert xy coordinates to pixel values."""
    idxs = X + np.array([w / 2, h / 2])
    idxs[:, 0] = w - idxs[:, 0]
    return (mag * idxs).astype(np.int32)


def to_ego_frame(
    coords: np.ndarray,
    ego_pose: np.ndarray,
    vertical: bool = True,
) -> np.ndarray:
    """
    Convert coordinates in the global frame to the ego's frame.

    If vertical then the frame is rotated so the ego is aligned
    vertically (pi/2).
    """
    return rotate_coords(
        coords - ego_pose[None, :2],
        -ego_pose[3] - (math.pi / 2 if vertical else 0.0),
    )



import os
import warnings
from contextlib import suppress
from copy import deepcopy
from typing import Dict, List, Optional, Type

import numpy as np
from lxml import etree

from scenario_gym.entity import Entity, Pedestrian, Vehicle
from scenario_gym.road_network import RoadNetwork
from scenario_gym.scenario import Scenario
from scenario_gym.trajectory import Trajectory

from .catalogs import read_catalog


def import_scenario(
    osc_file: str,
    relabel: bool = True,
    entity_types: Optional[List[Type[Entity]]] = None,
) -> Scenario:
    """
    Import a scenario from an OpenScenario file.

    Parameters
    ----------
    osc_file : str
        The filepath to the OpenScenario file.

    relabel : bool
        Whether to relabel entities after loading.

    entity_types : Optional[List[Type[Entity]]]
        Additional entity types to use when loading the scenario. Can be used to
        allow custom entities to be loaded from OpenSCENARIO.

    """
    if not os.path.exists(osc_file):
        raise FileNotFoundError

    cwd = os.path.dirname(osc_file)
    et = etree.parse(osc_file)
    osc_root = et.getroot()
    entities = {}

    # Read catalogs:
    catalogs: Dict[str, Dict[str, Entity]] = {}
    for catalog_location in osc_root.iterfind("CatalogLocations/"):
        rel_catalog_path = catalog_location.find("Directory").attrib["path"]
        catalog_path = os.path.join(cwd, rel_catalog_path)
        for catalog_file in os.listdir(catalog_path):
            if catalog_file.endswith(".xosc"):
                catalog, entries = read_catalog(
                    os.path.join(catalog_path, catalog_file),
                    relative_catalog_path=rel_catalog_path,
                    entity_types=entity_types,
                )
                catalogs[catalog.catalog_name] = entries

    # Import road network:
    scene_graph_file = osc_root.find("RoadNetwork/SceneGraphFile")
    if scene_graph_file is not None:
        rn_path = scene_graph_file.attrib["filepath"]
    else:
        logic_file = osc_root.find("RoadNetwork/LogicFile")
        rn_path = logic_file.attrib["filepath"]

    filepath = os.path.join(cwd, rn_path)
    extension = os.path.splitext(filepath)[1]
    if extension == "":
        filepath = f"{filepath}.json"

    if not os.path.exists(filepath):
        warnings.warn(f"Could not find road network file: {filepath}.")

    if extension in (".json", ""):
        road_network = RoadNetwork.create_from_json(filepath)
    elif extension == ".xodr":
        road_network = RoadNetwork.create_from_xodr(filepath)
    else:
        road_network = None

    # add the entities to the scenario
    for scenario_object in osc_root.iterfind("Entities/ScenarioObject"):
        entity_ref = scenario_object.attrib["name"]
        cat_ref = scenario_object.find("CatalogReference")
        if cat_ref is None:
            raise NotImplementedError(
                "Scenario objects can only be loaded from catalog references."
            )
        catalog_name = cat_ref.attrib["catalogName"]
        entry_name = cat_ref.attrib["entryName"]
        try:
            entity = deepcopy(catalogs[catalog_name][entry_name])
            entity.ref = entity_ref
            entities[entity_ref] = entity
        except KeyError as e:
            if catalog_name not in catalogs:
                warnings.warn(f"Could not find catalog: {catalog_name}")
            elif entry_name not in catalogs[catalog_name]:
                warnings.warn(
                    f"Could not find entry {entry_name} in catalog {catalog_name}."
                )
            else:
                raise e

    # Read init actions:
    for private in osc_root.iterfind("Storyboard/Init/Actions/Private"):
        entity_ref = private.attrib["entityRef"]
        for wp in private.iterfind(
            "PrivateAction/TeleportAction/Position/WorldPosition"
        ):
            tp = traj_point_from_time_and_position(0, wp)
            # Add a single-waypoint trajectory:
            if entity_ref in entities:
                entities[entity_ref].trajectory = Trajectory(np.stack([tp], axis=0))

    # Read maneuver actions:
    for maneuver_group in osc_root.iterfind("Storyboard/Story/Act/ManeuverGroup"):
        entity_ref = maneuver_group.find("Actors/EntityRef")
        assert (
            entity_ref is not None
        ), "Could not find entity reference in maneuver group."
        entity_ref = entity_ref.attrib["entityRef"]
        trajectory_points = []

        vertices = maneuver_group.findall(
            "Maneuver/Event/Action/PrivateAction/RoutingAction/"
            + "FollowTrajectoryAction/TrajectoryRef/Trajectory/Shape/"
            + "Polyline/Vertex"
        )
        vertices.extend(
            maneuver_group.findall(
                "Maneuver/Event/Action/PrivateAction/RoutingAction/"
                + "FollowTrajectoryAction/Trajectory/Shape/Polyline/Vertex"
            )
        )
        for vertex in vertices:
            t = float(vertex.attrib["time"])
            wp = vertex.find("Position/WorldPosition")
            trajectory_points.append(traj_point_from_time_and_position(t, wp))
        if entity_ref in entities:
            traj_data = np.stack(trajectory_points, axis=0)
            if (np.isnan(traj_data[:, 3]).sum() > 0) and (road_network is not None):
                traj_data[:, 3] = road_network.elevation_at_point(
                    traj_data[:, 1], traj_data[:, 2]
                )
            entities[entity_ref].trajectory = Trajectory(traj_data)

    scenario = Scenario(
        list(entities.values()),
        name=os.path.splitext(os.path.basename(osc_file))[0],
        path=osc_file,
        road_network=road_network,
    )

    if relabel:
        scenario = relabel_scenario(scenario)

    return scenario


def relabel_scenario(scenario: Scenario) -> Scenario:
    """
    Relabel the entities of the scenario.

    Will be relabelled to ego, vehicle_1, vehicle_2,
    ..., pedestrian_1, ..., other_1, ...

    """
    vehicles, pedestrians, others = 0, 0, 0
    scenario.entities[0].ref = "ego"
    old_to_new = {}
    for e in scenario.entities[1:]:
        cur = e.ref
        with suppress(KeyError):
            scenario._ref_to_entity.pop(cur)
        if isinstance(e, Vehicle):
            e.ref = f"vehicle_{vehicles}"
            vehicles += 1
        elif isinstance(e, Pedestrian):
            e.ref = f"pedestrian_{pedestrians}"
            pedestrians += 1
        else:
            e.ref = f"other_{others}"
            others += 1
        scenario._ref_to_entity[e.ref] = e
        old_to_new[cur] = e.ref
    for action in scenario.actions:
        action.entity_ref = old_to_new[action.entity_ref]
    return scenario


def traj_point_from_time_and_position(t, world_position) -> np.ndarray:
    """Return the trajectory point as an array [t, x, y, z, h, p, r]."""
    return np.array(
        [
            t,
            float(world_position.attrib["x"]),
            float(world_position.attrib["y"]),
            float(world_position.attrib.get("z", np.NaN)),
            float(world_position.attrib.get("h", np.NaN)),
            float(world_position.attrib.get("p", np.NaN)),
            float(world_position.attrib.get("r", np.NaN)),
        ],
    )



from scenario_gym.entity.base import Entity
from scenario_gym.entity.batch import BatchReplayEntity
from scenario_gym.entity.pedestrian import Pedestrian
from scenario_gym.entity.vehicle import Vehicle



from typing import List

import numpy as np
from shapely.geometry import LineString, Point

from scenario_gym.agent import Agent
from scenario_gym.entity import Entity
from scenario_gym.pedestrian.action import PedestrianAction
from scenario_gym.pedestrian.behaviour import PedestrianBehaviour
from scenario_gym.pedestrian.controller import PedestrianController
from scenario_gym.pedestrian.observation import PedestrianObservation
from scenario_gym.pedestrian.sensor import PedestrianSensor


class PedestrianAgent(Agent):
    """A pedestrian agent with a behaviour model."""

    def __init__(
        self,
        entity: Entity,
        route: List[np.array],
        speed_desired: float,
        behaviour: PedestrianBehaviour,
        max_speed: float = 5.0,
        head_rot_angle: float = 0.0,
        distance_threshold: float = 1.0,
    ):
        """Init the agent."""
        super().__init__(
            entity,
            PedestrianController(entity, max_speed=max_speed),
            PedestrianSensor(
                entity,
                head_rot_angle=head_rot_angle,
                distance_threshold=distance_threshold,
            ),
        )
        self.goal_idx = 0
        self.speed_desired = speed_desired
        self.behaviour = behaviour
        self.force = np.array([0.0, 0.0])

        self.route = route
        self.route_geom = LineString(route)
        self.route_arcs = np.hstack(
            [[0.0], np.linalg.norm(np.diff(route, axis=0), axis=1).cumsum()]
        )

    def _step(self, observation: PedestrianObservation) -> PedestrianAction:
        """
        Produce the next action.

        Parameters
        ----------
        observation : PedestrianObservation
            All info from environment within perception radius.

        """
        # Change goal point to next one in path of close enough
        if self.goal_idx <= len(self.route) - 1:
            s = self.route_geom.project(Point(*observation.pose[:2]))
            self.goal_idx = np.argwhere(self.route_arcs <= s).max() + 1
        if self.goal_idx <= len(self.route) - 1:
            speed, heading = self.behaviour.step(observation, self)
        else:  # reached goal
            speed = 0
            heading = 0
            self.force[:] = 0
        return PedestrianAction(speed, heading)
